/*
 *  IFF_ILBMHandler.cpp - Amiga IFF ILBM datatype handler
 *
 *  Written by Tim Stack (stack@eng.utah.edu),
 *  adapted by Edmund Vermeulen (edmundv@xs4all.nl).
 */

#include <DataHandler.h>
#include <Datatypes.h>
#include <DataStreams.h>
#include <InterfaceDefs.h>
#include <CheckBox.h>

#define ILBM_FORMAT 'ILBM'

char handlerName[] = "IFF_ILBMHandler";
char handlerInfo[] = "Written by Tim Stack & Edmund Vermeulen";
int32 handlerVersion = 110;

Format inputFormats[] =
{
	{ ILBM_FORMAT, DATA_BITMAP, 1.0 * 0.8, 0.9, "image/x-iff-ilbm", "Amiga IFF ILBM" },
	{ DATA_BITMAP, DATA_BITMAP, 0.9 * 1.0, 0.7, "image/x-be-bitmap", "Be Bitmap" },
	{ 0, 0, 0.0, 0.0, "", "" }
};

Format outputFormats[] =
{
	{ DATA_BITMAP, DATA_BITMAP, 0.9 * 1.0, 0.7, "image/x-be-bitmap", "Be Bitmap" },
	{ ILBM_FORMAT, DATA_BITMAP, 1.0 * 0.8, 0.9, "image/x-iff-ilbm", "Amiga IFF ILBM" },
	{ 0, 0, 0.0, 0.0, "", "" }
};

/* Amiga IFF stuff */
#define DATA_SIZE 12
#define mskNone 0
#define mskHasMask 1
#define cmpNone 0
#define cmpByteRun1 1
#define HAM 0x800

typedef char BYTE;
typedef uchar UBYTE;
typedef int16 WORD;
typedef uint16 UWORD;

typedef UBYTE Masking;
typedef UBYTE Compression;

typedef struct {
	UWORD w, h;
	WORD x, y;
	UBYTE nPlanes;
	Masking masking;
	Compression compression;
	UBYTE pad1;
	UWORD transparentColor;
	UBYTE xAspect, yAspect;
	WORD pageWidth, pageHeight;
} BitmapHeader;

typedef struct {
	UBYTE red;
	UBYTE green;
	UBYTE blue;
} ColorRegister;

/* settings stuff */
#define CHANGE_COMPRESSION 'SWCO'

static Compression compression_setting = cmpByteRun1;

class IFFConfigView : public BCheckBox
{
public:
	IFFConfigView(void);
	virtual void AttachedToWindow(void);
	virtual void MessageReceived(BMessage *msg);
};

IFFConfigView::IFFConfigView(void)
	: BCheckBox(BRect(0.0, 0.0, 90.0, 17.0), "", "Compression", new BMessage(CHANGE_COMPRESSION))
{
	if (compression_setting == cmpByteRun1)
		SetValue(B_CONTROL_ON);
}

void IFFConfigView::AttachedToWindow(void)
{
	SetTarget(this);
	BCheckBox::AttachedToWindow();
}

void IFFConfigView::MessageReceived(BMessage *msg)
{
	switch (msg->what)
	{
		case CHANGE_COMPRESSION:
			if (Value() == B_CONTROL_ON)
				compression_setting = cmpByteRun1;
			else
				compression_setting = cmpNone;
			break;

		default:
			BCheckBox::MessageReceived(msg);
			break;
	}
}

/* parse the IFF file */
static bool process_chunks( DStream &inStream, DStream &outStream )
{
	int32 lpc;
	uint32 chunk_name;
	bool dead = false, done = false;
	int32 chunk_size, size;
	ColorRegister *cmap = NULL;
	DATABitmap hdr;
	BitmapHeader bmh;
	char data[DATA_SIZE];
	bool ham = false;
	bool camg_found = false, bmhd_found = false;

	/* how much initial stuff do we need.  FORM(size)ILBM so its 12 */
	size = DATA_SIZE;
	if( (inStream.Read( data, size ) != B_NO_ERROR) || (size != DATA_SIZE) )
		return false;

	/* check if its really IFF ILBM */
	if( (strncmp( data, "FORM", 4 ) == 0) && (strncmp( data + 8, "ILBM", 4 ) == 0) )
	{
		hdr.magic = DATA_BITMAP;
		hdr.colors = B_RGB_32_BIT;  /* always convert to true color */

		/* allocate space for 256 colors */
		cmap = (ColorRegister *)malloc( 256 * sizeof( ColorRegister ) );
		if( !cmap )
			dead = true;

		uint32 name_size = 4;
		/* we loop so long as there hasn't been some error (!dead),
		   the bitmap hasn't been written (!done),
		   there wasn't an error in reading from inStream,
		   and we could read four bytes */
		while( !dead &&
		       !done &&
		       (inStream.Read( &chunk_name, name_size ) == B_NO_ERROR) &&
		       (name_size == 4) )
		{
			size = 4;
			/* we've read the chunk name, now we need to get the size of the chunk */
			if( inStream.Read( &chunk_size, size ) == B_NO_ERROR )
			{
				switch( chunk_name )
				{
					case 'BMHD':  /* bitmap header */
						bmhd_found = true;
						size = sizeof( BitmapHeader );
						if( inStream.Read( &bmh, size ) == B_NO_ERROR )
						{
							hdr.bounds.Set( 0.0, 0.0, bmh.w - 1.0, bmh.h - 1.0 );
							/* check if bitplane depth is supported */
							if( (bmh.nPlanes < 0) || ((bmh.nPlanes > 8) && (bmh.nPlanes != 24)) )
								dead = true;
							/* check if we support the compression method */
							if( (bmh.compression != cmpNone) && (bmh.compression != cmpByteRun1) )
								dead = true;
							/* using the depth we need to get the number of row bytes and
							   compute the total size of the data */
							hdr.rowBytes = bmh.w * 4;
							hdr.dataSize = bmh.h * hdr.rowBytes;
						}
						else
							dead = true;
						break;

					case 'CMAP':  /* colormap */
						/* read in the colormap */
						if( (chunk_size <= 256 * 3) && (inStream.Read( cmap, chunk_size) == B_NO_ERROR) )
						{
							/* if the lower four bits of all color map entries are unused,
							   we assume that they are old style 4 bit color values
							   that need to be converted to 8 bit */
							bool four_bit_color = true;
							for( lpc = 0; lpc < (chunk_size / 3); lpc++ )
							{
								if( (cmap[lpc].red   & 0x0f) ||
								    (cmap[lpc].green & 0x0f) ||
								    (cmap[lpc].blue  & 0x0f) )
								{
									four_bit_color = false;
									break;
								}
							}
							/* convert the colormap to 8 bit if it already isn't.
							   we have to duplicate the value in the low order bits
							   so that white is 0xff instead of 0xf0 */
							if( four_bit_color )
							{
								for( lpc = 0; lpc < chunk_size / 3; lpc++ )
								{
									cmap[lpc].red   = cmap[lpc].red   | (cmap[lpc].red   >> 4);
									cmap[lpc].green = cmap[lpc].green | (cmap[lpc].green >> 4);
									cmap[lpc].blue  = cmap[lpc].blue  | (cmap[lpc].blue  >> 4);
								}
							}
							/* if there are 32 colors and 6 planes then we assume its ehb */
							if( (chunk_size / 3 == 32) && (bmh.nPlanes == 6) )
							{
								for( lpc = 32; lpc < 64; lpc++ )
								{
									/* fill in the higher order table elements with
									half brite colors */
									cmap[lpc].red   = cmap[lpc - 32].red   / 2;
									cmap[lpc].green = cmap[lpc - 32].green / 2;
									cmap[lpc].blue  = cmap[lpc - 32].blue  / 2;
								}
							}
						}
						else
							dead = true;
						break;

					case 'CAMG':  /* Commodore Amiga modeid */
						camg_found = true;
						uint32 mode_id;
						if( (chunk_size == 4) && (inStream.Read( &mode_id, chunk_size ) == B_NO_ERROR) )
							ham = ((mode_id & HAM) != 0);
						else
							dead = true;
						break;

					case 'BODY':  /* the bitplane data */
						if( bmhd_found )
						{
							unsigned char *body = NULL, *row = NULL, *row32 = NULL;
							unsigned char *place, *row_place, *row32_place, *chunk_end;
							int32 curr_row, curr_pix, curr_plane, row_len, skip, n, count;
							unsigned char plane_bit;
							
							if( !camg_found && (bmh.nPlanes == 6) )
								ham = true;
	
							row_len = ((bmh.w + 15) / 16) * 2;
	
							/* the default format for storing 24 bit data in IFF
							   is a little brain dead.  There are 24 planes where
							   the first plane is r7 then r6, r5, r4, and so on.
							   Then it will go onto green and blue. So when we
							   put it into the 32 bit row data we have to skip
							   over other color bytes. */
							if( bmh.nPlanes == 24 )
							{
								skip = 4;
							}
							else
							{
								skip = 1;
								row = (unsigned char *)malloc( row_len * 8 ); /* chunky pixel row */
								if( !row )
									dead = true;
							}
	
							row32 = (unsigned char *)malloc( row_len * 32 + 4); /* 32 bit pixel row */
							if( !row32 )
								dead = true;
							body = (unsigned char *)malloc( chunk_size ); /* the whole body data read in */
							if( !body )
								dead = true;
							place = body;
							chunk_end = body + chunk_size;
	
							if( !dead &&
							    (inStream.Read( body, chunk_size ) == B_NO_ERROR) &&
							    (outStream.Write( &hdr, sizeof( DATABitmap ) ) == B_NO_ERROR) )
							{
								if( ham )
								{
									/* if its ham we have to read the planes in regularly and
									   then convert them to a 32bit row which we will write out.
									   if the first pixel in a row is a modify then it
									   modifies the border pixel which is the 0 color
									   register so we need to set this up instead of
									   writing a special case in the ham -> 32 bit code */
									row32[0] = cmap[0].blue;
									row32[1] = cmap[0].green;
									row32[2] = cmap[0].red;
									row32[3] = 0;
									row32 += 4;  /* restore this before freeing! */
								}
	
								/* loop through every row in the picture */
								for( curr_row = 0; (curr_row < bmh.h) && !dead; curr_row++ )
								{
									/* clean out the last row we read in */
									if( bmh.nPlanes == 24 )
										memset( row32, 0, row_len * 32 );
									else
										memset( row, 0, row_len * 8 );
	
									/* planes are interleaved which is groovy so we can read in the planes
									   convert to a chunky row and then write them out */
									for( curr_plane = 0; curr_plane < bmh.nPlanes + (bmh.masking == mskHasMask); curr_plane++ )
									{
										/* a 32 bit BBitmap has the colors ordered blue green red alpha,
										   instead of red green blue */
										if( bmh.nPlanes == 24 )
											row_place = row32 + 2 - (curr_plane / 8);
										else
											row_place = row;
	
										/* current bit in the chunky pixel we're reading in */
										plane_bit = (1 << (curr_plane % 8));
										/* check for compression */
										if( bmh.compression == cmpNone )
										{
											/* check that we don't read past the buffer */
											if( place + row_len <= chunk_end )
											{
												for( lpc = 0 ; lpc < row_len; lpc++ )
												{
													/* don't decode the masking bitplane if there is one */
													if( curr_plane < bmh.nPlanes )
													{
														for( curr_pix = 1 << 7; curr_pix > 0; curr_pix >>= 1 )
														{
															/* check if theres a pixel in the bitplane
															   and OR it in the chunky pixel */
															if( curr_pix & *place )
															{
																*row_place |= plane_bit;
															}
															row_place += skip;
														}
													}
													place++;
												}
											}
										}
										else
										{
											/* its compressed so we need to unpack, its just a simple
											   RLE packer so this is simple */
											count = row_len;
											while( (count > 0) && (place < chunk_end))
											{
												n = (signed char) *place;
												place++;
												if( n >= 0 )
												{
													n = n + 1;
													if( n > count )
														n = count;
													if( place + n > chunk_end )
														n = chunk_end - place;
													count -= n;
													while( n > 0 )
													{
														if( curr_plane < bmh.nPlanes )
														{
															for( curr_pix = 1 << 7; curr_pix > 0; curr_pix >>= 1 )
															{
																if( curr_pix & *place )
																{
																	*row_place |= plane_bit;
																}
																row_place += skip;
															}
														}
														place++;
														n--;
													}
												}
												else
													if( n != -128 )
													{
														n = -n + 1;
														if( n > count )
															n = count;
														if( place >= chunk_end )
															n = 0;
														count -= n;
														if( curr_plane < bmh.nPlanes )
														{
															while( n > 0 )
															{
																for( curr_pix = 1 << 7; curr_pix > 0; curr_pix >>= 1 )
																{
																	if( curr_pix & *place )
																	{
																		*row_place |= plane_bit;
																	}
																	row_place += skip;
																}
																n--;
															}
														}
														place++;
													}
											}
										}
									}
	
									row_place = row;
									row32_place = row32;
	
									if( ham )
									{
										/* it's hold and modify mode; only modify one color component and
										   hold the other ones, or look up the color in the color map.
										   both ham6 and ham8 are supported */
										for( lpc = 0; lpc < bmh.w; lpc++ )
										{
											switch( *row_place & (bmh.nPlanes == 8 ? 0xC0 : 0x30) )
											{
												/* lookup color */
												case 0x00:
													*(row32_place++) = cmap[*row_place].blue;
													*(row32_place++) = cmap[*row_place].green;
													*(row32_place++) = cmap[*row_place].red;
													break;
												/* modify blue, hold the rest */
												case 0x10:
												case 0x40:
													if( bmh.nPlanes == 8 )
														*(row32_place++) = (*(row32_place - 4) & 0x03) | ((*row_place & 0x3f) << 2);
													else
														*(row32_place++) = (*row_place & 0x0f) | ((*row_place & 0x0f) << 4);
													*(row32_place++) = *(row32_place - 4);
													*(row32_place++) = *(row32_place - 4);
													break;
												/* modify red */
												case 0x20:
												case 0x80:
													*(row32_place++) = *(row32_place - 4);
													*(row32_place++) = *(row32_place - 4);
													if( bmh.nPlanes == 8 )
														*(row32_place++) = (*(row32_place - 4) & 0x03) | ((*row_place & 0x3f) << 2);
													else
														*(row32_place++) = (*row_place & 0x0f) | ((*row_place & 0x0f) << 4);
													break;
												/* modify green */
												case 0x30:
												case 0xC0:
													*(row32_place++) = *(row32_place - 4);
													if( bmh.nPlanes == 8 )
														*(row32_place++) = (*(row32_place - 4) & 0x03) | ((*row_place & 0x3f) << 2);
													else
														*(row32_place++) = (*row_place & 0x0f) | ((*row_place & 0x0f) << 4);
													*(row32_place++) = *(row32_place - 4);
													break;
											}
											*(row32_place++) = 0;  /* alpha, not used */
											row_place++;
										}
									}
									else
									{
										if( bmh.nPlanes != 24 )
										{
											/* convert chunky 8 bit to 32 bit */
											for( lpc = 0; lpc < bmh.w; lpc++ )
											{
												*(row32_place++) = cmap[*row_place].blue;
												*(row32_place++) = cmap[*row_place].green;
												*(row32_place++) = cmap[*row_place].red;
												*(row32_place++) = 0;
												row_place++;
											}
										}
									}
									size = bmh.w * 4;
									if( outStream.Write( row32, size ) != B_NO_ERROR )
										dead = true;
								}
							}
							else
								dead = true;
	
							/* free some memory */
							if( body )
								free( body );
							if( row )
								free( row );
							if( row32 )
							{
								if( ham )
									row32 -= 4;  /* restore mem pointer */
								free( row32 );
							}
						}
						else
							dead = true;

						done = true;  /* no need to process any further chunks */
						break;

					default:  /* unknown chunk */
						inStream.Seek( inStream.Position() + chunk_size );
						break;
				}
				if( chunk_size % 2 == 1 )  /* uneven chunk size */
					inStream.Seek( inStream.Position() + 1 );
			}
		}
		if( cmap )
			free( cmap );
	}
	return( done && !dead );
}

/* RLE encodes the data in runs of a maximum of 128 bytes, preceded by a
   control byte that indicates the number of bytes in the run.
   A positive control byte is followed by n+1 (literal) bytes of data,
   a negative control byte is followed by the byte to be repeated -n+1 times,
   -128 is a no-op. */
static int32 run_length_encode(char *source, char *dest, int32 in_size)
{
	int32 same_count = 0;
	int32 run_count = 0;
	bool dump_mode = true;
	bool are_same;
	char last_char;

	char *in_source = source;
	char *source_end = source + in_size;
	char *in_dest = dest;

	while (in_source < source_end)
	{
		if (dump_mode)
		{
			if ((same_count > 0) && (*in_source == last_char))
			{
				same_count++;
			}
			else
			{
				last_char = *in_source;
				same_count = 1;
			}

			if (same_count == 3)
			{
				if (run_count > 2)
				{
					*(in_dest++) = run_count - 3;
					memcpy(in_dest, in_source - run_count, run_count - 2);
					in_dest += run_count - 2;
				}

				same_count = 2;
				dump_mode = false;
			}
			else
			{
				run_count++;
				in_source++;
				if ((run_count == 128) || (in_source == source_end))
				{
					*(in_dest++) = run_count - 1;
					memcpy(in_dest, in_source - run_count, run_count);
					in_dest += run_count;

					same_count = 0;
					run_count = 0;
				}
			}
		}
		else
		{
			are_same = (*in_source == last_char);
			if (are_same)
			{
				same_count++;
				in_source++;
			}

			if (!are_same || (same_count == 128) || (in_source == source_end))
			{
				*(in_dest++) = - same_count + 1;
				*(in_dest++) = last_char;

				same_count = 0;
				run_count = 0;
				dump_mode = true;
			}
		}
	}
	return in_dest - dest;
}

// Convert DATA_BITMAP to IFF ILBM 
static bool write_iff(DStream &inStream, DStream &outStream)
{
	bool alive = false;
	DATABitmap bmp;
	BitmapHeader bmh;
	int32 size, skip, chunks;

	size = sizeof(DATABitmap);
	if ((inStream.Read(&bmp, size) == B_NO_ERROR) &&
	    (size == sizeof(DATABitmap)) &&
	    (bmp.magic == DATA_BITMAP) &&
	    bmp.bounds.IsValid())
	{
		if ((bmp.colors == B_RGB_32_BIT) || (bmp.colors == B_COLOR_8_BIT))
		{
			// Fill in the ILBM bitmap header
			bmh.w = (UWORD) bmp.bounds.right - (UWORD) bmp.bounds.left + 1;
			bmh.h = (UWORD) bmp.bounds.bottom - (UWORD) bmp.bounds.top + 1;
			bmh.x = 0;
			bmh.y = 0;
			switch (bmp.colors)
			{
				case B_RGB_32_BIT:
					bmh.nPlanes = 24;
					chunks = 2;
					skip = 4;
					break;

				case B_COLOR_8_BIT:
					bmh.nPlanes = 8;
					chunks = 0;
					skip = 1;
					break;
			}
			bmh.masking = mskNone;
			bmh.compression = compression_setting;
			bmh.pad1 = 0;
			bmh.transparentColor = 0;
			bmh.xAspect = 1;
			bmh.yAspect = 1;
			bmh.pageWidth = bmh.w;
			bmh.pageHeight = bmh.h;

			// Some rigorous checking
			if ((bmp.rowBytes == ((bmh.w * skip + 3) / 4) * 4) &&
			    (bmp.dataSize == bmp.rowBytes * bmh.h))
			{
				unsigned char *line_buffer, *plane_buffer, *rle_buffer = NULL;

				// We want even lines of bitplane data
				int32 row_len = ((bmh.w + 15) / 16) * 2;

				// We must be prepared for a worst case scenario, when RLE actually
				// increases the data size, instead of making it smaller.
				int32 rle_row_len = row_len + ((row_len + 127) / 128);

				if (((line_buffer = new unsigned char[bmp.rowBytes]) != NULL) &&
				    ((plane_buffer = new unsigned char[row_len]) != NULL) &&
				    ((bmh.compression == cmpNone) ||
				     ((rle_buffer = new unsigned char[rle_row_len]) != NULL)))
				{
					alive = true;

					// IFF identifier
					size = 4;
					outStream.Write("FORM", size);
					// IFF data size
					int32 base_size = 20 + sizeof(BitmapHeader);
					if (bmp.colors == B_COLOR_8_BIT)
						base_size += 776;  // for the CMAP chunk
					int32 size_val = base_size + row_len * bmh.nPlanes * bmh.h;
					size = 4;
					outStream.Write(&size_val, size);
					// ILBM & BMHD identifiers
					size = 8;
					outStream.Write("ILBMBMHD", size);
					// BMHD chunk size
					size_val = sizeof(BitmapHeader);
					size = 4;
					outStream.Write(&size_val, size);
					// BMHD chunk data
					size = sizeof(BitmapHeader);
					outStream.Write(&bmh, size);
					// 24-bit doesn't need a colour map
					if (bmp.colors == B_COLOR_8_BIT)
					{
						// Copy Be's colour map to our own
						ColorRegister *cmap = new ColorRegister[256];
						const rgb_color *be_map = system_colors()->color_list;
						for (int32 i = 0; i < 256; i++)
						{
							cmap[i].red   = be_map[i].red;
							cmap[i].green = be_map[i].green;
							cmap[i].blue  = be_map[i].blue;
						}
						// CMAP identifier
						size = 4;
						outStream.Write("CMAP", size);
						// CMAP chunk size
						size_val = 768;
						size = 4;
						outStream.Write(&size_val, size);
						// CMAP chunk data
						size = 768;
						outStream.Write(cmap, size);

						delete[] cmap;
					}
					// BODY identifier
					size = 4;
					outStream.Write("BODY", size);
					// BODY chunk size
					int32 body_pos = outStream.Position();
					size_val = row_len * bmh.nPlanes * bmh.h;
					size = 4;
					outStream.Write(&size_val, size);

					// In case of compression, we will keep track of the real body
					// size written, and Seek() back later to write the correct IFF
					// and BODY chunk sizes.
					int32 rle_body_size = 0;

					// The bitplane data has to be interleaved for each pixel row,
					// so we will go through it one line at the time
					for (UWORD y = 0; (y < bmh.h) && alive; y++)
					{
						size = bmp.rowBytes;
						if ((inStream.Read(line_buffer, size) != B_NO_ERROR) || (size != bmp.rowBytes))
							alive = false;

						// Chunky to planar conversion; 8-bit just has one chunk,
						// but a 32-bit bitmap has 3 chunks (red, green and blue)
						for (int32 c = chunks; (c >= 0) && alive; c--)
						{
							unsigned char plane_mask = 1;
							// Create 8 bitplanes
							for (int32 b = 0; (b < 8) && alive; b++)
							{
								unsigned char *in_plane_buffer = plane_buffer;
								// Do all pixels of one row
								for (UWORD x = 0; x < row_len * 8; x += 8)
								{
									unsigned char plane_byte = 0;
									int32 place = c + x * skip;

									// Let's do 8 bits in one go!
									if ((x     < bmh.w) && (line_buffer[place           ] & plane_mask)) plane_byte |= 0x80;
									if ((x + 1 < bmh.w) && (line_buffer[place +     skip] & plane_mask)) plane_byte |= 0x40;
									if ((x + 2 < bmh.w) && (line_buffer[place + 2 * skip] & plane_mask)) plane_byte |= 0x20;
									if ((x + 3 < bmh.w) && (line_buffer[place + 3 * skip] & plane_mask)) plane_byte |= 0x10;
									if ((x + 4 < bmh.w) && (line_buffer[place + 4 * skip] & plane_mask)) plane_byte |= 0x08;
									if ((x + 5 < bmh.w) && (line_buffer[place + 5 * skip] & plane_mask)) plane_byte |= 0x04;
									if ((x + 6 < bmh.w) && (line_buffer[place + 6 * skip] & plane_mask)) plane_byte |= 0x02;
									if ((x + 7 < bmh.w) && (line_buffer[place + 7 * skip] & plane_mask)) plane_byte |= 0x01;

									// Write a byte of bitplane data
									*(in_plane_buffer++) = plane_byte;
								}
								plane_mask <<= 1;

								if (bmh.compression == cmpNone)
								{
									size = row_len;
									if (outStream.Write(plane_buffer, size) != B_NO_ERROR)
										alive = false;
								}
								else
								{
									size = run_length_encode((char *) plane_buffer, (char *) rle_buffer, row_len);
									rle_body_size += size;
									if (outStream.Write(rle_buffer, size) != B_NO_ERROR)
										alive = false;
								}
							}
						}
					}
					if ((bmh.compression == cmpByteRun1) && alive)
					{
						// Uneven chunks have to be padded
						int32 pad = (rle_body_size & 1);
						if (pad)
						{
							char pad_char = '\0';
							size = 1;
							if (outStream.Write(&pad_char, size) != B_NO_ERROR)
								alive = false;
						}
						if (alive)
						{
							// Write new IFF and BODY chunk size
							int32 size_val = base_size + rle_body_size + pad;
							size = 4;
							if ((outStream.Seek(4) != B_NO_ERROR) ||
							    (outStream.Write(&size_val, size) != B_NO_ERROR) || (size != 4) ||
							    (outStream.Seek(body_pos) != B_NO_ERROR) ||
							    (outStream.Write(&rle_body_size, size) != B_NO_ERROR))
							{
								alive = false;
							}
						}
					}
				}
				if (line_buffer)
					delete[] line_buffer;
			    if (plane_buffer)
			    	delete[] plane_buffer;
			    if (rle_buffer)
			    	delete[] rle_buffer;
			}
		}
	}
	return alive;
}

/* Identify function */
int32
Identify(
	DStream &			inSource,
	const Format *		inFormat,
	BMessage *			ioExtension,
	DATAInfo &			outInfo,
	uint32				outType)
{
	if( outType && (outType != DATA_BITMAP) && (outType != ILBM_FORMAT) )
		return DATA_NO_HANDLER;

	char data[32];
	int32 size = 32;
	if( (inSource.Read( data, size ) != B_NO_ERROR) || (size != 32) )
		return DATA_NO_HANDLER;

	/* check if it is IFF ILBM */
	if( (strncmp( data, "FORM", 4 ) == 0) && (strncmp( data + 8, "ILBM", 4 ) == 0) )
	{
		inFormat = &inputFormats[0];
	}
	else
	{
		/* check if it is DATA_BITMAP */
		DATABitmap *hdr = (DATABitmap *) data;
		if( (hdr->magic == DATA_BITMAP) &&
		    hdr->bounds.IsValid() &&
		    (hdr->rowBytes >= 4) && ((hdr->rowBytes & 3) == 0) &&
		    ((hdr->colors == B_RGB_32_BIT) || (hdr->colors == B_COLOR_8_BIT)) &&
		    (hdr->dataSize == ((int32) hdr->bounds.bottom - (int32) hdr->bounds.top + 1) * hdr->rowBytes) )
		{
			inFormat = &inputFormats[1];
		}
		else
		{
			return DATA_NO_HANDLER;
		}
	}

	/* set the DATAInfo to the recognized type */
	outInfo.formatType = inFormat->type;
	outInfo.formatGroup = inFormat->t_cls;
	outInfo.formatQuality = inFormat->quality;
	outInfo.formatCapability = inFormat->capability;
	strcpy( outInfo.formatName, inFormat->formatName );
	strcpy( outInfo.MIMEName, inFormat->MIME );
	return B_NO_ERROR;
}

/* Translate function */
int32
Translate(
	DStream &			inStream,
	const DATAInfo &	inInfo,
	BMessage *			ioExtension,
	uint32				outFormat,
	DStream &			outStream)
{
	if( outFormat == 0 )
		outFormat = DATA_BITMAP;

	if( (inInfo.formatType == ILBM_FORMAT) && (outFormat == DATA_BITMAP) )
	{
		if( process_chunks( inStream, outStream ) )
			return B_NO_ERROR;
	}
	if( (inInfo.formatType == DATA_BITMAP) && (outFormat == ILBM_FORMAT) )
	{
		if( write_iff( inStream, outStream ) )
			return B_NO_ERROR;
	}
	return DATA_NO_HANDLER;
}

/* MakeConfig function */
int32
MakeConfig(
	BMessage *	ioExtension,
	BView * & 	outView,
	BRect & 	outExtent)
{
	outView = new IFFConfigView();
	outExtent = outView->Bounds();
	return B_NO_ERROR;
}
